/*
 * Copyright (C) 2009-2010 National Marrow Donor Program. All Rights Reserved.
 */

package net.agnis.metadatatool.ui.swing

import groovy.swing.SwingBuilder
import java.awt.BorderLayout
import java.awt.Color
import java.awt.Component
import java.awt.Cursor
import java.awt.Dimension
import java.awt.Toolkit
import javax.swing.BoxLayout
import javax.swing.JButton
import javax.swing.JCheckBox
import javax.swing.JDialog
import javax.swing.JFileChooser
import javax.swing.JOptionPane
import javax.swing.JPasswordField
import javax.swing.JTabbedPane
import javax.swing.JTable
import javax.swing.JTextArea
import javax.swing.JTextField
import javax.swing.ListSelectionModel
import javax.swing.ProgressMonitor
import javax.swing.RootPaneContainer
import javax.swing.SwingConstants
import javax.swing.UIManager
import javax.swing.WindowConstants
import java.util.prefs.Preferences
import net.agnis.cadsrapi41.CurationRetrieverFactory
import net.agnis.cadsrapi41.POIHelper
import net.agnis.cadsrapi41.RetrieverFactory
import net.agnis.cadsrapi41.WSClientHelper
import org.apache.commons.logging.Log
import org.apache.commons.logging.LogFactory

class MetadataConsole
{
    static Log log = LogFactory.getLog(MetadataConsole.class)

    Boolean removeDraftNewFormTab = true
    Boolean removeCdeReviewTab = true
    Boolean hideNonProtocolReleasedForms = true
    Boolean expandValueDomain = false

    static private prefs = Preferences.userNodeForPackage(MetadataConsole)

    Boolean releasedFormTableRefreshInProgress = false
    Boolean draftNewFormTableRefreshInProgress = false
    Boolean cdeReviewTableRefreshInProgress = false

    SwingBuilder swing
    RootPaneContainer frame
    JTable releasedFormTable
    JTable draftNewFormTable
    JTable cdeReviewTable
    WSClientHelper wsClientHelper
    JTabbedPane tabbedPane
    List releasedFormList
    List draftNewFormList
    def cdeReviewClassificationScheme

    File currentFileChooserDir = new File(prefs.get('currentFileChooserDir', '.'))
    Boolean useProxyServer = prefs.getBoolean('useProxyServer', false)
    String proxyHost = prefs.get('proxyHost', '')
    String proxyPort = prefs.get('proxyPort', '')
    String proxyUser = prefs.get('proxyUser', '')
    String proxyPassword = prefs.get('proxyPassword', '')
    Boolean useHttpBasicAuthentication = prefs.getBoolean('useHttpBasicAuthentication', false)
    String httpUser = prefs.get('httpUser', '')
    String httpPassword = prefs.get('httpPassword', '')

    static void main(args)
    {
        UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        def console = new MetadataConsole()
        console.run()
    }

    MetadataConsole()
    {
        wsClientHelper = new WSClientHelper(
            this.class.getResource(
                '/cadsrapi41Service_WSDL/customized/cadsrapi41Service.wsdl'),
            this.useProxyServer,
            this.proxyHost,
            this.proxyPort,
            this.proxyUser,
            this.proxyPassword,
            this.useHttpBasicAuthentication,
            this.httpUser,
            this.httpPassword)
    }

    def loadMetadataConsoleProperties()
    {
        Properties p = new Properties()
        p.load(this.class.getResourceAsStream('/MetadataConsole.properties'))
        this.removeDraftNewFormTab =
            !"false".equalsIgnoreCase(p.get('removeDraftNewFormTab'))
        this.removeCdeReviewTab =
            !"false".equalsIgnoreCase(p.get('removeCdeReviewTab'))
        this.hideNonProtocolReleasedForms =
            !"false".equalsIgnoreCase(p.get('hideNonProtocolReleasedForms'))
        this.expandValueDomain =
            "true".equalsIgnoreCase(p.get('expandValueDomain'))
    }

    def getVersionBuildString()
    {
        Properties p = new Properties()
        p.load(this.class.getResourceAsStream('/VersionInfo.properties'))
        return 'Version:  ' + p.get('version') + '_' + p.get('buildNumber')
    }

    def displayAbout()
    {
        def pane = swing.optionPane()
        pane.setMessage('Welcome to the AGNIS\u00ae Metadata Tool!\n' +
            getVersionBuildString())
        def dialog = pane.createDialog(frame, 'About Metadata Tool')
        dialog.show()
    }

    def checkFileExistsOverwrite(parent, File file)
    {
        if(!file) return false
        if(file.exists())
        {
            def answer = JOptionPane.showConfirmDialog(
                parent,
                'File exists: ' + file.name + '\nOverwrite?',
                'Continue?',
                JOptionPane.YES_NO_OPTION)
            if(answer != JOptionPane.YES_OPTION) return false
            file.delete()
        }
        return true
    }

    def checkFileReadable(Component parent, File file)
    {
        if(!file) return false
        if(file.canRead()) return true
        JOptionPane.showMessageDialog(
            parent,
            'Unable to read file\n' + file.canonicalPath,
            'Error',
            JOptionPane.ERROR_MESSAGE)
        return false
    }

    def setCursor(Component component, int cursor)
    {
        component.setCursor(Cursor.getPredefinedCursor(cursor));
    }

    def hideAndDisposeDialog(JDialog dialog)
    {
        dialog.hide()
        dialog.dispose()
    }

    def computeFormNumber(longName)
    {
        def matcher = (longName =~ /^\d+/)
        if(matcher)
        {
            return matcher[0]
        }
        return ''
    }

    def computePopupLocation(int width, int height)
    {
        def x = (int)((frame.size.width - width) / 2 + frame.location.x)
        def y = (int)((frame.size.height - height) / 2 + frame.location.y)
        def screenSize = Toolkit.defaultToolkit.screenSize
        if(x > screenSize.width) x = (int)(screenSize.width - 100)
        if(x < 0) x = 0
        if(y > screenSize.height) y = (int)(screenSize.height - 100)
        if(y < 0) y = 0
        return [x,y]
    }

    def computePopupLocation(Dimension d)
    {
        computePopupLocation((int)d.width, (int)d.height)
    }

    String replaceExtension(String filename, String extension)
    {
        if(!filename)
        {
            return null
        }

        def parts = filename.split(/\./)
        if((parts.size() > 1) && (parts[-1] ==~ /[\w\d]+/))
        {
            return parts[0..-2].join('.') + extension 
        }
        return filename + extension
    }

    ProgressMonitor openDownloadProgressMonitor(
        String filename, int progressCount)
    {
        if(!progressCount) progressCount = 500
        def mon = new ProgressMonitor(frame, 'Downloading caDSR XML',
            filename, 0, progressCount + (int)(progressCount/100) + 3)
        mon.millisToPopup = 0
        return mon
    }

    def exit()
    {
        if (frame instanceof java.awt.Window)
        {
            frame.hide()
            frame.dispose()
        }
        System.exit(0)
    }

    def refreshReleasedFormTable()
    {
        synchronized(releasedFormTableRefreshInProgress)
        {
            if(releasedFormTableRefreshInProgress)
            {
                return
            }
            releasedFormTableRefreshInProgress = true
        }

        setCursor(releasedFormTable.parent, Cursor.WAIT_CURSOR);
        Thread.start
        {
            def error = null
            try
            {
                populateReleasedFormTable([])
                releasedFormList = wsClientHelper.retrieveReleasedFormList(
                    this.hideNonProtocolReleasedForms)
                def data = []
                releasedFormList.each
                {
                    crf ->
                    data << new Vector([ crf.longName, crf.publicID, crf.version])
                }
                populateReleasedFormTable(data)
            }
            catch(Throwable t)
            {
                log.error('Unexpected error.', t)
                error = t
            }
            finally
            {
                setCursor(releasedFormTable.parent, Cursor.DEFAULT_CURSOR);
                releasedFormTableRefreshInProgress = false
            }

            if(error)
            {
                JOptionPane.showMessageDialog(frame,
                    error.class.name + ': ' + error.message,
                    'Error during refresh',
                    JOptionPane.ERROR_MESSAGE)
            }
        }
    }

    def refreshDraftNewFormTable()
    {
        synchronized(draftNewFormTableRefreshInProgress)
        {
            if(draftNewFormTableRefreshInProgress)
            {
                return
            }
            draftNewFormTableRefreshInProgress = true
        }

        setCursor(draftNewFormTable.parent, Cursor.WAIT_CURSOR);
        Thread.start
        {
            def error = null
            try
            {
                populateDraftNewFormTable([])
                draftNewFormList = wsClientHelper.retrieveDraftNewFormList()
                def data = []
                draftNewFormList.each
                {
                    crf ->
                    data << new Vector([ crf.longName, crf.publicID, crf.version])
                }
                populateDraftNewFormTable(data)
            }
            catch(Throwable t)
            {
                log.error('Unexpected error.', t)
                error = t
            }
            finally
            {
                setCursor(draftNewFormTable.parent, Cursor.DEFAULT_CURSOR);
                draftNewFormTableRefreshInProgress = false
            }

            if(error)
            {
                JOptionPane.showMessageDialog(frame,
                    error.class.name + ': ' + error.message,
                    'Error during refresh',
                    JOptionPane.ERROR_MESSAGE)
            }
        }
    }

    def populateReleasedFormTable(data)
    {
        populateFormTable(releasedFormTable, data)
    }

    def populateDraftNewFormTable(data)
    {
        populateFormTable(draftNewFormTable, data)
    }

    def populateFormTable(table, data)
    {
        def columnNames =
        [
            'Case Report Form (CRF) Long Name',
            'CRF Public ID',
            'CRF Version'
        ]
        table.model = new UneditableTableModel(
            new Vector(data), new Vector(columnNames))
        def columnModel = table.columnModel
        columnModel.getColumn(0).preferredWidth = 300
        columnModel.getColumn(1).preferredWidth = 150
        columnModel.getColumn(2).preferredWidth = 150
        columnModel.columnSelectionAllowed = false
        table.selectionModel.selectionMode =
            ListSelectionModel.SINGLE_SELECTION
    }

    def downloadReleasedFormXML()
    {
        displayCrfXmlDownloadWizard(releasedFormTable, releasedFormList)
    }

    def downloadDraftNewFormXML()
    {
        displayCrfXmlDownloadWizard(draftNewFormTable, draftNewFormList)
    }

    def downloadFormXML(String crfId, File xmlFile, File translationFile, File reportFile)
    {
        def count = wsClientHelper.retrieveFormQuestionCount(crfId)
        ProgressMonitor progressMonitor = openDownloadProgressMonitor(xmlFile.name, count)
        Thread.start
        {
            def retrieved
            def dlError = null
            def marshallingError = null
            def translationFileError = null
            def reportFileError = null
            def completed = false

            try
            {
                retrieved = wsClientHelper.retrieveForm(crfId, progressMonitor, expandValueDomain)
            }
            catch(Throwable t)
            {
                log.error('Unexpected error.', t)
                dlError = t
            }

            if(!dlError && !progressMonitor.isCanceled())
            {
                try
                {
                    wsClientHelper.marshal(retrieved, 'Form', xmlFile)

                    POIHelper ph = new POIHelper(wsClientHelper)

                    if(translationFile)
                    {
                        try
                        {
                            def wb = ph.createCrfTranslationWorkbook(retrieved)
                            ph.writeExcelFile(wb, translationFile)
                        }
                        catch(Throwable t)
                        {
                            log.error('Unexpected error.', t)
                            translationFileError = t
                        }
                    }

                    if(reportFile)
                    {
                        try
                        {
                            def wb = ph.createCrfReportWorkbook(retrieved)
                            ph.writeExcelFile(wb, reportFile)
                        }
                        catch(Throwable t)
                        {
                            log.error('Unexpected error.', t)
                            reportFileError = t
                        }
                    }

                    completed = true
                }
                catch(Throwable t)
                {
                    log.error('Unexpected error.', t)
                    marshallingError = t
                }
            }
            progressMonitor.close()

            if(dlError)
            {
                JOptionPane.showMessageDialog(frame,
                    dlError.class.name + ': ' + dlError.message,
                    'Error downloading ' + xmlFile.name,
                    JOptionPane.ERROR_MESSAGE)
            }
            else if(progressMonitor.isCanceled())
            {
                JOptionPane.showMessageDialog(frame,
                    xmlFile.name,
                    'Download Cancelled',
                    JOptionPane.INFORMATION_MESSAGE)
            }
            else if(marshallingError)
            {
                JOptionPane.showMessageDialog(frame,
                    marshallingError.class.name + ': ' + marshallingError.message,
                    'Error creating ' + xmlFile.name,
                    JOptionPane.ERROR_MESSAGE)
            }
            else if(translationFileError)
            {
                JOptionPane.showMessageDialog(frame,
                    translationFileError.class.name + ': ' + translationFileError.message,
                    'Error generating ' + translationFile.name,
                    JOptionPane.ERROR_MESSAGE)
            }
            else if(reportFileError)
            {
                JOptionPane.showMessageDialog(frame,
                    reportFileError.class.name + ': ' + reportFileError.message,
                    'Error generating ' + reportFile.name,
                    JOptionPane.ERROR_MESSAGE)
            }
            else if(completed)
            {
                def msg = 'Created: ' + xmlFile.name
                if(translationFile) msg += '\nGenerated: ' + translationFile.name
                if(reportFile) msg += '\nGenerated: ' + reportFile.name

                JOptionPane.showMessageDialog(frame,
                    msg,
                    'Operation completed',
                    JOptionPane.INFORMATION_MESSAGE)
            }
        }
    }

    def refreshCdeReviewTable()
    {
        synchronized(cdeReviewTableRefreshInProgress)
        {
            if(cdeReviewTableRefreshInProgress)
            {
                return
            }
            cdeReviewTableRefreshInProgress = true
        }

        setCursor(cdeReviewTable.parent, Cursor.WAIT_CURSOR);
        Thread.start
        {
            def error = null
            try
            {
                populateCdeReviewTable([])
                cdeReviewClassificationScheme = wsClientHelper.retrieveCdeReviewClassificationScheme()
                def data = []
                cdeReviewClassificationScheme.classSchemeClassSchemeItemCollection.item.each
                {
                    def csi = it.classificationSchemeItem
                    data << new Vector([ csi.longName, csi.publicID, csi.version])
                }
                populateCdeReviewTable(data)
            }
            catch(Throwable t)
            {
                log.error('Unexpected error.', t)
                error = t
            }
            finally
            {
                setCursor(cdeReviewTable.parent, Cursor.DEFAULT_CURSOR);
                cdeReviewTableRefreshInProgress = false
            }

            if(error)
            {
                JOptionPane.showMessageDialog(frame,
                    error.class.name + ': ' + error.message,
                    'Error during refresh',
                    JOptionPane.ERROR_MESSAGE)
            }
        }
    }

    def populateCdeReviewTable(data)
    {
        def columnNames =
        [
            'Classification Scheme Item (CSI) Long Name',
            'CSI Public ID',
            'CSI Version'
        ]
        cdeReviewTable.model = new UneditableTableModel(
            new Vector(data), new Vector(columnNames))
        def columnModel = cdeReviewTable.columnModel
        columnModel.getColumn(0).preferredWidth = 300
        columnModel.getColumn(1).preferredWidth = 150
        columnModel.getColumn(2).preferredWidth = 150
        columnModel.columnSelectionAllowed = false
        cdeReviewTable.selectionModel.selectionMode =
            ListSelectionModel.SINGLE_SELECTION
    }

    def downloadCdeReviewXML()
    {
        def sel = cdeReviewTable.selectionModel.anchorSelectionIndex
        if(sel < 0)
        {
            JOptionPane.showMessageDialog(frame,
                'Nothing selected!',
                'Error',
                JOptionPane.ERROR_MESSAGE)
            return
        }
        def cscsi = cdeReviewClassificationScheme.classSchemeClassSchemeItemCollection.item[sel]
        def csi = cscsi.classificationSchemeItem
        def formNumber = computeFormNumber(csi.longName)
        if(formNumber) formNumber += '_'
        def defaultFilename = formNumber + 'CSI_' + csi.publicID + '_' + csi.version + '.xml'
        File outputFile = selectFile('Save', defaultFilename)
        frame.update(frame.graphics)
        if(outputFile && checkFileExistsOverwrite(frame, outputFile))
        {
            setCursor(frame, Cursor.WAIT_CURSOR);
            frame.update(frame.graphics)
            try
            {
                def count = wsClientHelper.retrieveCscsiItemCount(cscsi.id)
                ProgressMonitor progressMonitor = openDownloadProgressMonitor(outputFile.name, count)
                Thread.start
                {
                    def retrieved
                    def error = null
                    def completed = false

                    try
                    {
                        retrieved = wsClientHelper.retrieveClassSchemeClassSchemeItem(cscsi.id, progressMonitor)
                    }
                    catch(Throwable t)
                    {
                        log.error('Unexpected error.', t)
                        error = t
                    }

                    if(!error && !progressMonitor.isCanceled())
                    {
                        try
                        {
                            wsClientHelper.marshal(retrieved, 'ClassSchemeClassSchemeItem', outputFile)
                            completed = true
                        }
                        catch(Throwable t)
                        {
                            log.error('Unexpected error.', t)
                            error = t
                        }
                    }
                    progressMonitor.close()

                    if(error)
                    {
                        JOptionPane.showMessageDialog(frame,
                            error.class.name + ': ' + error.message,
                            'Error downloading ' + outputFile.name,
                            JOptionPane.ERROR_MESSAGE)
                    }
                    else if(progressMonitor.isCanceled())
                    {
                        JOptionPane.showMessageDialog(frame,
                            outputFile.name,
                            'Download Cancelled',
                            JOptionPane.INFORMATION_MESSAGE)
                    }
                    else if(completed)
                    {
                        JOptionPane.showMessageDialog(frame,
                            outputFile.name,
                            'Download Completed',
                            JOptionPane.INFORMATION_MESSAGE)
                    }
                }
            }
            finally
            {
                setCursor(frame, Cursor.DEFAULT_CURSOR);
            }
        }
    }

    File selectFile(title, defaultFilename)
    {
        selectFile(frame, title, defaultFilename, false)
    }

    File selectFile(parent, title, defaultFilename, isCanonicalFilename)
    {
        def fc = new JFileChooser()
        fc.fileSelectionMode = JFileChooser.FILES_ONLY
        fc.acceptAllFileFilterUsed = true

        if(!defaultFilename)
        {
            fc.selectedFile = new File(currentFileChooserDir, '.')
        }
        else if(isCanonicalFilename)
        {
            fc.selectedFile = new File(defaultFilename)
        }
        else
        {
            fc.selectedFile = new File(currentFileChooserDir, defaultFilename)
        }

        if (fc.showDialog(parent, title) == JFileChooser.APPROVE_OPTION)
        {
            currentFileChooserDir = fc.currentDirectory
            prefs.put('currentFileChooserDir', currentFileChooserDir.path)
            return fc.selectedFile
        }
        else
        {
            return null
        }
    }

    def displayProxySettingsDialog()
    {
        JDialog dialog
        JCheckBox useProxyServerCheckbox, useHttpBasicAuthenticationCheckbox
        JTextField proxyHostField, proxyPortField, proxyUserField, httpUserField
        JPasswordField proxyPasswordField, httpPasswordField

        dialog = swing.dialog(
            title: 'Proxy Settings',
            size: [300,300],
            pack: true,
            resizable: true,
            defaultCloseOperation: WindowConstants.DISPOSE_ON_CLOSE)
        {
            borderLayout()
            panel(constraints: CENTER, border: emptyBorder(20,30,0,30))
            {
                gridBagLayout()
                useProxyServerCheckbox =
                    checkBox(text: 'Use HTTP Proxy Server',
                        selected: useProxyServer,
                        constraints: gbc(gridx: 1, gridy: 0, anchor: WEST))
                label('Host name:   ',
                    constraints: gbc(gridx: 0, gridy: 1, anchor: EAST))
                proxyHostField =
                    textField(columns: 15,  text: proxyHost,
                        enabled: useProxyServer,
                        constraints: gbc(gridx: 1, gridy: 1, anchor: WEST))
                label('Port:   ',
                    constraints: gbc(gridx: 0, gridy: 2, anchor: EAST))
                proxyPortField =
                    textField(columns: 15, text: proxyPort,
                        enabled: useProxyServer,
                        constraints: gbc(gridx: 1, gridy: 2, anchor: WEST))
                label('User name:   ',
                    constraints: gbc(gridx: 0, gridy: 3, anchor: EAST))
                proxyUserField =
                    textField(columns: 15, text: proxyUser,
                        enabled: useProxyServer,
                        constraints: gbc(gridx: 1, gridy: 3, anchor: WEST))
                label('Password:   ',
                    constraints: gbc(gridx: 0, gridy: 4, anchor: EAST))
                proxyPasswordField =
                    passwordField(columns: 15, text: proxyPassword,
                        enabled: useProxyServer,
                        constraints: gbc(gridx: 1, gridy: 4, anchor: WEST))

                useHttpBasicAuthenticationCheckbox =
                    checkBox(text: 'Use HTTP Basic Authentication',
                        selected: useHttpBasicAuthentication,
                        constraints: gbc(gridx: 1, gridy: 5, insets: [15,0,0,0], anchor: WEST))
                label('User name:   ',
                    constraints: gbc(gridx: 0, gridy: 6, anchor: EAST))
                httpUserField =
                    textField(columns: 15, text: httpUser,
                        enabled: useHttpBasicAuthentication,
                        constraints: gbc(gridx: 1, gridy: 6, anchor: WEST))
                label('Password:   ',
                    constraints: gbc(gridx: 0, gridy: 7, anchor: EAST))
                httpPasswordField =
                    passwordField(columns: 15, text: httpPassword,
                        enabled: useHttpBasicAuthentication,
                        constraints: gbc(gridx: 1, gridy: 7, anchor: WEST))
            }
            panel(constraints: SOUTH, border: emptyBorder(25,30,15,30))
            {
                boxLayout(axis: BoxLayout.LINE_AXIS)
                hbox
                {
                    hglue()
                    button(text: ' Save ', actionPerformed:
                        {
                            saveProxySettings(
                                dialog,
                                useProxyServerCheckbox,
                                proxyHostField,
                                proxyPortField,
                                proxyUserField,
                                proxyPasswordField,
                                useHttpBasicAuthenticationCheckbox,
                                httpUserField,
                                httpPasswordField)
                        })
                    hstrut(width: 10)
                    button(text: 'Cancel', actionPerformed:
                        {
                            hideAndDisposeDialog(dialog)
                        })
                }
            }
        }

        useProxyServerCheckbox.addItemListener(
            new CheckboxCheckedListener(useProxyServerCheckbox,
                [proxyHostField, proxyPortField, proxyUserField, proxyPasswordField]))
        useHttpBasicAuthenticationCheckbox.addItemListener(
            new CheckboxCheckedListener(useHttpBasicAuthenticationCheckbox,
                [httpUserField, httpPasswordField]))

        dialog.location = computePopupLocation(dialog.size)
        dialog.show()
    }

    def saveProxySettings(
        JDialog dialog,
        JCheckBox useProxyServerCheckbox,
        JTextField proxyHostField,
        JTextField proxyPortField,
        JTextField proxyUserField,
        JPasswordField proxyPasswordField,
        JCheckBox useHttpBasicAuthenticationCheckbox,
        JTextField httpUserField,
        JPasswordField httpPasswordField)
    {
        if(useProxyServerCheckbox.selected)
        {
            def error = null
            Integer portNum
            try
            {
                portNum = Integer.parseInt(proxyPortField.text)
            }
            catch(NumberFormatException e)
            {
                error = e
            }

            if(error || (portNum < 0) || (portNum > 65535))
            {
                JOptionPane.showMessageDialog(frame,
                    'Proxy server port must be an integer\n between 0 and 65535.',
                    'Error',
                    JOptionPane.ERROR_MESSAGE)
                return
            }
        }
        // read input values
        useProxyServer = useProxyServerCheckbox.selected
        proxyHost = proxyHostField.text
        proxyPort = proxyPortField.text
        proxyUser = proxyUserField.text
        proxyPassword = new String(proxyPasswordField.password)
        useHttpBasicAuthentication = useHttpBasicAuthenticationCheckbox.selected
        httpUser = httpUserField.text
        httpPassword = new String(httpPasswordField.password)
        // save settings
        prefs.putBoolean('useProxyServer', useProxyServer)
        prefs.put('proxyHost', proxyHost)
        prefs.put('proxyPort', proxyPort)
        prefs.put('proxyUser', proxyUser)
        prefs.put('proxyPassword', proxyPassword)
        prefs.putBoolean('useHttpBasicAuthentication', useHttpBasicAuthentication)
        prefs.put('httpUser', httpUser)
        prefs.put('httpPassword', httpPassword)

        hideAndDisposeDialog(dialog)

        JOptionPane.showMessageDialog(frame,
            'New proxy settings will take effect\n after the next program restart.',
            'Information',
            JOptionPane.INFORMATION_MESSAGE)
    }

    def displayCrfXmlDownloadWizard(JTable table, List formList)
    {
        def wizDialogList = []
        def ddesc

        def sel = table.selectionModel.anchorSelectionIndex
        if(sel < 0)
        {
            JOptionPane.showMessageDialog(frame,
                'Nothing selected!',
                'Error',
                JOptionPane.ERROR_MESSAGE)
            return
        }
        def crf = formList[sel]

        ddesc = new WizardDialogDescription()
        ddesc.windowTitle = 'Download CRF XML'
        ddesc.description =
            crf.longName + '\n' +
            'CRF Public ID ' + crf.publicID + ', Version ' + crf.version + '\n\n' +
            'Specify output file for Case Report Form metadata ' +
            '(cadsrapi41 XML format).'
        ddesc.context.crf = crf
        def formNumber = computeFormNumber(crf.longName)
        if(formNumber) formNumber += '_'
        ddesc.inputValue = new File(currentFileChooserDir, formNumber + 'CRF_' + crf.publicID + '_' + crf.version + '.xml').canonicalPath
        ddesc.beforeNextAction =
        {
            int index, List<WizardDialogDescription> wizardDialogDescriptionList ->

            assert index == 0
            try
            {
                def xmlFile = new File(wizardDialogDescriptionList[0].inputValue)
                if(!wizardDialogDescriptionList[1].inputValue)
                {
                    wizardDialogDescriptionList[1].inputValue =
                        replaceExtension(wizardDialogDescriptionList[0].inputValue, '_report.xls')
                }
            }
            catch(Throwable t)
            {
                log.error('Unexpected error.', t)
                JOptionPane.showMessageDialog(
                    wizardDialogDescriptionList[index].dialog,
                    t.class.name + ': ' + t.message,
                    'Error',
                    JOptionPane.ERROR_MESSAGE)
                return false
            }

            return true
        }
        wizDialogList << ddesc

        ddesc = new WizardDialogDescription()
        ddesc.windowTitle = 'Download CRF XML'
        ddesc.description =
            'The CRF Report Spreadsheet is similar to the spreadsheet ' +
            'generated by the Form Builder tool, but includes additional ' +
            'information required by AGNIS\u00ae (Microsoft\u00ae ' +
            'Excel\u00ae format).'
        ddesc.checkboxLabel = 'Generate CRF Report Spreadsheet'
        ddesc.beforeNextAction =
        {
            int index, List<WizardDialogDescription> wizardDialogDescriptionList ->

            assert index == 1
            try
            {
                def xmlFile = new File(wizardDialogDescriptionList[0].inputValue)
                if(wizardDialogDescriptionList[1].checkboxSelected)
                {
                    def reportFile = new File(wizardDialogDescriptionList[1].inputValue)
                }
                if(!wizardDialogDescriptionList[2].inputValue)
                {
                    wizardDialogDescriptionList[2].inputValue =
                        replaceExtension(wizardDialogDescriptionList[0].inputValue, '_translation.xls')
                }
            }
            catch(Throwable t)
            {
                log.error('Unexpected error.', t)
                JOptionPane.showMessageDialog(
                    wizardDialogDescriptionList[index].dialog,
                    t.class.name + ': ' + t.message,
                    'Error',
                    JOptionPane.ERROR_MESSAGE)
                return false
            }

            return true
        }
        wizDialogList << ddesc

        ddesc = new WizardDialogDescription()
        ddesc.windowTitle = 'Download CRF XML'
        ddesc.description =
            'The CRF Translation Spreadsheet includes separate workbooks ' +
            'for field and value mappings, to facilitate translation ' +
            'between AGNIS\u00ae and other systems (Microsoft\u00ae ' +
            'Excel\u00ae format).'
        ddesc.checkboxLabel = 'Generate CRF Translation Spreadsheet'
        ddesc.finishAction =
        {
            int index, List<WizardDialogDescription> wizardDialogDescriptionList ->

            File xmlFile, reportFile, translationFile
            assert index == 2
            JDialog dialog = wizardDialogDescriptionList[index].dialog
            try
            {
                xmlFile = new File(wizardDialogDescriptionList[0].inputValue)
                if(!checkFileExistsOverwrite(dialog, xmlFile))
                    return false
                if(wizardDialogDescriptionList[1].checkboxSelected)
                {
                    reportFile = new File(wizardDialogDescriptionList[1].inputValue)
                    if(!checkFileExistsOverwrite(dialog, reportFile))
                        return false
                }
                if(wizardDialogDescriptionList[2].checkboxSelected)
                {
                    translationFile = new File(wizardDialogDescriptionList[2].inputValue)
                    if(!checkFileExistsOverwrite(dialog, translationFile))
                        return false
                }
            }
            catch(Throwable t)
            {
                log.error('Unexpected error.', t)
                JOptionPane.showMessageDialog(
                    wizardDialogDescriptionList[index].dialog,
                    t.class.name + ': ' + t.message,
                    'Error',
                    JOptionPane.ERROR_MESSAGE)
                return false
            }

            def error = null

            setCursor(frame, Cursor.WAIT_CURSOR)
            setCursor(dialog, Cursor.WAIT_CURSOR)
            frame.update(frame.graphics)
            dialog.update(dialog.graphics)
            try
            {
                downloadFormXML(
                    wizardDialogDescriptionList[0].context.crf.id,
                    xmlFile,
                    translationFile,
                    reportFile)
                return true
            }
            catch(Throwable t)
            {
                log.error('Unexpected error.', t)
                error = t
            }
            finally
            {
                setCursor(frame, Cursor.DEFAULT_CURSOR);
                setCursor(dialog, Cursor.DEFAULT_CURSOR);
            }

            if(error)
            {
                JOptionPane.showMessageDialog(
                    wizardDialogDescriptionList[index].dialog,
                    error.class.name + ': ' + error.message,
                    'Error',
                    JOptionPane.ERROR_MESSAGE)
            }

            return false
        }
        wizDialogList << ddesc

        displayWizardDialog(0, wizDialogList)
    }

    def displayCrfReportWizard()
    {
        def wizDialogList = []
        def ddesc

        ddesc = new WizardDialogDescription()
        ddesc.windowTitle = 'Generate CRF Report Spreadsheet'
        ddesc.description = 'Case Report Form (CRF) XML File \n\n' +
            'Select a file containing CRF metadata ' +
            'downloaded from the caDSR (cadsrapi41 XML format).'
        ddesc.beforeNextAction =
        {
            int index, List<WizardDialogDescription> wizardDialogDescriptionList ->

            try
            {
                def inputFile = new File(wizardDialogDescriptionList[0].inputValue)
                if(!checkFileReadable(wizardDialogDescriptionList[index].dialog, inputFile))
                    return false
                if(!wizardDialogDescriptionList[1].inputValue)
                {
                    wizardDialogDescriptionList[1].inputValue =
                        replaceExtension(wizardDialogDescriptionList[0].inputValue, '_report.xls')
                }
            }
            catch(Throwable t)
            {
                log.error('Unexpected error.', t)
                JOptionPane.showMessageDialog(
                    wizardDialogDescriptionList[index].dialog,
                    t.class.name + ': ' + t.message,
                    'Error',
                    JOptionPane.ERROR_MESSAGE)
                return false
            }

            return true
        }
        wizDialogList << ddesc

        ddesc = new WizardDialogDescription()
        ddesc.windowTitle = 'Generate CRF Report Spreadsheet'
        ddesc.description =
            'The CRF Report Spreadsheet is similar to the spreadsheet ' +
            'generated by the Form Builder tool, but includes additional ' +
            'information required by AGNIS\u00ae (Microsoft\u00ae ' +
            'Excel\u00ae format).'
        ddesc.finishAction =
        {
            int index, List<WizardDialogDescription> wizardDialogDescriptionList ->

            JDialog dialog = wizardDialogDescriptionList[index].dialog
            File inputFile, outputFile
            try
            {
                inputFile = new File(wizardDialogDescriptionList[0].inputValue)
                if(!checkFileReadable(wizardDialogDescriptionList[index].dialog, inputFile))
                    return false
                outputFile = new File(wizardDialogDescriptionList[1].inputValue)
                if(!outputFile || !checkFileExistsOverwrite(dialog, outputFile))
                {
                    return false
                }
            }
            catch(Throwable t)
            {
                log.error('Unexpected error.', t)
                JOptionPane.showMessageDialog(
                    wizardDialogDescriptionList[index].dialog,
                    t.class.name + ': ' + t.message,
                    'Error',
                    JOptionPane.ERROR_MESSAGE)
                return false
            }

            def error = null

            setCursor(frame, Cursor.WAIT_CURSOR)
            setCursor(dialog, Cursor.WAIT_CURSOR)
            frame.update(frame.graphics)
            dialog.update(dialog.graphics)
            try
            {
                POIHelper p = new POIHelper(wsClientHelper)
                def form = p.readCRFFile(inputFile)
                if(!form || (form.class.name != 'gov.nih.nci.cadsr.domain.jaxb.Form'))
                {
                    JOptionPane.showMessageDialog(
                        wizardDialogDescriptionList[index].dialog,
                        'Unable to read CRF data from file ' + inputFile.name,
                        'Error',
                        JOptionPane.ERROR_MESSAGE)
                }
                else
                {
                    def wb = p.createCrfReportWorkbook(form)
                    p.writeExcelFile(wb, outputFile)
                    JOptionPane.showMessageDialog(
                        wizardDialogDescriptionList[index].dialog,
                        'Output file:  ' + outputFile.name,
                        'CRF Report Spreadsheet Created',
                        JOptionPane.INFORMATION_MESSAGE)

                    return true
                }
            }
            catch(Throwable t)
            {
                log.error('Unexpected error.', t)
                error = t
            }
            finally
            {
                setCursor(frame, Cursor.DEFAULT_CURSOR);
                setCursor(dialog, Cursor.DEFAULT_CURSOR);
            }

            if(error)
            {
                JOptionPane.showMessageDialog(frame,
                    error.class.name + ': ' + error.message,
                    'Error',
                    JOptionPane.ERROR_MESSAGE)
            }

            return false
        }
        wizDialogList << ddesc

        displayWizardDialog(0, wizDialogList)
    }

    def displayCrfTranslationWizard()
    {
        def wizDialogList = []
        def ddesc

        ddesc = new WizardDialogDescription()
        ddesc.windowTitle = 'Generate CRF Translation Spreadsheet'
        ddesc.description = 'Case Report Form (CRF) XML File \n\n' +
            'Select a file containing CRF metadata ' +
            'downloaded from the caDSR (cadsrapi41 XML format).'
        ddesc.beforeNextAction =
        {
            int index, List<WizardDialogDescription> wizardDialogDescriptionList ->

            try
            {
                def inputFile = new File(wizardDialogDescriptionList[0].inputValue)
                if(!checkFileReadable(wizardDialogDescriptionList[index].dialog, inputFile))
                    return false
                if(!wizardDialogDescriptionList[1].inputValue)
                {
                    wizardDialogDescriptionList[1].inputValue =
                        replaceExtension(wizardDialogDescriptionList[0].inputValue, '_translation.xls')
                }
            }
            catch(Throwable t)
            {
                log.error('Unexpected error.', t)
                JOptionPane.showMessageDialog(
                    wizardDialogDescriptionList[index].dialog,
                    t.class.name + ': ' + t.message,
                    'Error',
                    JOptionPane.ERROR_MESSAGE)
                return false
            }

            return true
        }
        wizDialogList << ddesc

        ddesc = new WizardDialogDescription()
        ddesc.windowTitle = 'Generate CRF Translation Spreadsheet'
        ddesc.description =
            'The CRF Translation Spreadsheet includes separate workbooks ' +
            'for field and value mappings, to facilitate translation ' +
            'between AGNIS\u00ae and other systems (Microsoft\u00ae ' +
            'Excel\u00ae format).'
        ddesc.finishAction =
        {
            int index, List<WizardDialogDescription> wizardDialogDescriptionList ->

            JDialog dialog = wizardDialogDescriptionList[index].dialog
            File inputFile, outputFile
            try
            {
                inputFile = new File(wizardDialogDescriptionList[0].inputValue)
                if(!checkFileReadable(wizardDialogDescriptionList[index].dialog, inputFile))
                    return false
                outputFile = new File(wizardDialogDescriptionList[1].inputValue)
                if(!outputFile || !checkFileExistsOverwrite(dialog, outputFile))
                {
                    return false
                }
            }
            catch(Throwable t)
            {
                log.error('Unexpected error.', t)
                JOptionPane.showMessageDialog(
                    wizardDialogDescriptionList[index].dialog,
                    t.class.name + ': ' + t.message,
                    'Error',
                    JOptionPane.ERROR_MESSAGE)
                return false
            }

            def error = null

            setCursor(frame, Cursor.WAIT_CURSOR)
            setCursor(dialog, Cursor.WAIT_CURSOR)
            frame.update(frame.graphics)
            dialog.update(dialog.graphics)
            try
            {
                POIHelper p = new POIHelper(wsClientHelper)
                def form = p.readCRFFile(inputFile)
                if(!form || (form.class.name != 'gov.nih.nci.cadsr.domain.jaxb.Form'))
                {
                    JOptionPane.showMessageDialog(
                        wizardDialogDescriptionList[index].dialog,
                        'Unable to read CRF data from file ' + inputFile.name,
                        'Error',
                        JOptionPane.ERROR_MESSAGE)
                }
                else
                {
                    def wb = p.createCrfTranslationWorkbook(form)
                    p.writeExcelFile(wb, outputFile)
                    JOptionPane.showMessageDialog(
                        wizardDialogDescriptionList[index].dialog,
                        'Output file:  ' + outputFile.name,
                        'CRF Translation Spreadsheet Created',
                        JOptionPane.INFORMATION_MESSAGE)

                    return true
                }
            }
            catch(Throwable t)
            {
                log.error('Unexpected error.', t)
                error = t
            }
            finally
            {
                setCursor(frame, Cursor.DEFAULT_CURSOR);
                setCursor(dialog, Cursor.DEFAULT_CURSOR);
            }

            if(error)
            {
                JOptionPane.showMessageDialog(frame,
                    error.class.name + ': ' + error.message,
                    'Error',
                    JOptionPane.ERROR_MESSAGE)
            }

            return false
        }
        wizDialogList << ddesc

        displayWizardDialog(0, wizDialogList)
    }

    def displayWizardDialog(int index, List<WizardDialogDescription> wizardDialogDescriptionList)
    {
        WizardDialogDescription ddesc = wizardDialogDescriptionList[index]
        switch(ddesc.type)
        {
            case 'FILE_SELECTOR':
                displayFileSelectorWizardDialog(index, wizardDialogDescriptionList)
                break;

            default:
                throw new RuntimeException('Unexpected DialogDescription type: ' +
                    ddesc.type + '(index ' + index + ')')
        }
    }

    def displayFileSelectorWizardDialog(int index, List<WizardDialogDescription> wizardDialogDescriptionList)
    {
        WizardDialogDescription ddesc = wizardDialogDescriptionList[index]
        assert ddesc.type == 'FILE_SELECTOR'

        def dialog
        def inputField
        def browseButton
        def useFileCheckbox = null
        dialog = swing.dialog(
            title: ddesc.windowTitle,
            size: [350,250],
            pack: true,
            resizable: true,
            defaultCloseOperation: WindowConstants.DISPOSE_ON_CLOSE)
        {
            borderLayout()
            panel(constraints: EAST, border: emptyBorder(10,30,0,30))
            {
                gridBagLayout()
                def sizer = label(' ', preferredSize: [10,123],
                    constraints: gbc(gridheight: 2, gridx: 0, gridy: 0))
                def descriptionTextArea = textArea(text: ddesc.description,
                    focusable: false, editable: false, lineWrap: true,
                    wrapStyleWord: true, highlighter: null,
                    constraints: gbc(fill: BOTH, gridwidth: 2, gridx: 1, gridy: 0, insets: [15,0,20,0]))
                if(ddesc.checkboxLabel)
                {
                    useFileCheckbox = checkBox(text: ddesc.checkboxLabel,
                        selected: ddesc.checkboxSelected,
                        constraints: gbc(anchor: SOUTHWEST, gridwidth: 2, gridx: 1, gridy: 1, insets: [0,0,5,0]))
                }
                def inputLabel = label(text: ddesc.inputLabel,
                    constraints: gbc(anchor: EAST, gridx: 0, gridy: 2, insets: [0,0,0,10]))
                inputField = textField(text: ddesc.inputValue, columns: 20,
                    constraints: gbc(anchor: WEST, gridx: 1, gridy: 2))
                browseButton = button(text: 'Browse...',
                    constraints: gbc(anchor: WEST, gridx: 2, gridy: 2, insets: [0,10,0,0]),
                    actionPerformed:
                    {
                        def selected = selectFile(dialog, 'Open', inputField.text, true)
                        if(selected)
                        {
                            inputField.text = selected.canonicalPath
                        }
                    })
                descriptionTextArea.background = inputLabel.background
                descriptionTextArea.font = inputLabel.font
                descriptionTextArea.invalidate()
                // TODO: find a more robust, cross-platform method to control window size/layout
                sizer.preferredSize = new Dimension(10,
                        7 * descriptionTextArea.getFontMetrics(descriptionTextArea.font).height + 25)
                if(useFileCheckbox)
                {
                    inputField.enabled = useFileCheckbox.selected
                    browseButton.enabled = useFileCheckbox.selected
                }
            }
            panel(constraints: SOUTH, border: emptyBorder(20,30,15,30))
            {
                boxLayout(axis: BoxLayout.LINE_AXIS)
                hbox
                {
                    hglue()
                    if(index > 0)
                    {
                        button(text: '< Back', actionPerformed:
                            {
                                ddesc.inputValue = inputField.text
                                if(useFileCheckbox)
                                {
                                    ddesc.checkboxSelected = useFileCheckbox.selected
                                }
                                hideAndDisposeDialog(dialog)
                                wizardDialogDescriptionList[index].dialog = null
                                displayWizardDialog(index - 1, wizardDialogDescriptionList)
                            })
                        hstrut(width: 10)
                    }
                    if(index < wizardDialogDescriptionList.size() - 1)
                    {
                        button(text: 'Next >', actionPerformed:
                            {
                                ddesc.inputValue = inputField.text
                                if(useFileCheckbox)
                                {
                                    ddesc.checkboxSelected = useFileCheckbox.selected
                                }
                                if(ddesc.beforeNextAction.call(index, wizardDialogDescriptionList))
                                {
                                    hideAndDisposeDialog(dialog)
                                    wizardDialogDescriptionList[index].dialog = null
                                    displayWizardDialog(index + 1, wizardDialogDescriptionList)
                                }
                            })
                        hstrut(width: 10)
                    }
                    if(index == wizardDialogDescriptionList.size() - 1)
                    {
                        button(text: 'Finish', actionPerformed:
                            {
                                ddesc.inputValue = inputField.text
                                if(useFileCheckbox)
                                {
                                    ddesc.checkboxSelected = useFileCheckbox.selected
                                }
                                if(ddesc.finishAction.call(index, wizardDialogDescriptionList))
                                {
                                    hideAndDisposeDialog(dialog)
                                }
                            })
                        hstrut(width: 10)
                    }
                    button(text: 'Cancel', alignmentX: Component.RIGHT_ALIGNMENT,
                        actionPerformed:
                        {
                            hideAndDisposeDialog(dialog)
                        })
                }
            }
        }

        wizardDialogDescriptionList[index].dialog = dialog

        if(useFileCheckbox)
        {
            useFileCheckbox.addItemListener(
                new CheckboxCheckedListener(useFileCheckbox,
                    [inputField, browseButton]))
        }

        dialog.location = computePopupLocation(dialog.size)
        dialog.visible = true
    }

    def run()
    {
        swing = new SwingBuilder()
        frame = swing.frame(
            title: 'AGNIS\u00ae Metadata Tool',
            //location: [100,100],
            size: [600,400],
            defaultCloseOperation:WindowConstants.EXIT_ON_CLOSE)
        {
            try
            {
                current.locationByPlatform = true
            }
            catch(Exception e)
            {
                current.location = [100, 100]
            }

            menuBar
            {
                menu(text:'File', mnemonic: 'F')
                {
                    menuItem()
                    {
                        action(name: 'Proxy Settings...', closure: { displayProxySettingsDialog() })
                    }
                    menuItem()
                    {
                        action(name: 'Exit', closure: { exit() })
                    }
                }

                menu(text:'Report', mnemonic: 'R')
                {
                    menuItem()
                    {
                        action(name: 'CRF Report Spreadsheet...', closure: { displayCrfReportWizard() })
                    }
                    menuItem()
                    {
                        action(name: 'CRF Translation Spreadsheet...', closure: { displayCrfTranslationWizard() })
                    }
                }

                menu(text:'Help', mnemonic: 'H')
                {
                    menuItem()
                    {
                        action(name: 'About Metadata Tool', closure: { displayAbout() })
                    }
                }
            }

            tabbedPane = tabbedPane()
            {
                panel(name: 'Available Forms', border: emptyBorder(10,10,10,10))
                {
                    borderLayout()
                    panel(constraints: NORTH, border: emptyBorder(0,0,10,0))
                    {
                        boxLayout(axis: BoxLayout.X_AXIS)
                        hbox
                        {
                            button(text: 'Refresh List', actionPerformed:
                            {
                                refreshReleasedFormTable()
                            })
                        }
                    }
                    scrollPane(constraints: CENTER)
                    {
                        releasedFormTable = table()
                    }
                    panel(constraints: SOUTH, border: emptyBorder(10,0,0,0))
                    {
                        boxLayout(axis: BoxLayout.X_AXIS)
                        hbox
                        {
                            button(text: 'Download XML', actionPerformed:{ downloadReleasedFormXML() })
                        }
                    }
                }

                panel(name: 'Draft New Forms', border: emptyBorder(10,10,10,10))
                {
                    borderLayout()
                    panel(constraints: NORTH, border: emptyBorder(0,0,10,0))
                    {
                        boxLayout(axis: BoxLayout.X_AXIS)
                        hbox
                        {
                            button(text: 'Refresh List', actionPerformed:
                            {
                                refreshDraftNewFormTable()
                            })
                        }
                    }
                    scrollPane(constraints: CENTER)
                    {
                        draftNewFormTable = table()
                    }
                    panel(constraints: SOUTH, border: emptyBorder(10,0,0,0))
                    {
                        boxLayout(axis: BoxLayout.X_AXIS)
                        hbox
                        {
                            button(text: 'Download XML', actionPerformed:{ downloadDraftNewFormXML() })
                        }
                    }
                }

                panel(name: 'CDEs to Review', border: emptyBorder(10,10,10,10))
                {
                    borderLayout()
                    panel(constraints: NORTH, border: emptyBorder(0,0,10,0))
                    {
                        boxLayout(axis:BoxLayout.X_AXIS)
                        hbox
                        {
                            button(text: 'Refresh List', actionPerformed:
                                { refreshCdeReviewTable() })
                        }
                    }
                    scrollPane(constraints: CENTER)
                    {
                        cdeReviewTable = table()
                    }
                    panel(constraints: SOUTH, border: emptyBorder(10,0,0,0))
                    {
                        boxLayout(axis: BoxLayout.X_AXIS)
                        hbox
                        {
                            button(text: 'Download XML', actionPerformed: { downloadCdeReviewXML() })
                        }
                    }
                }
            }
        }

        populateReleasedFormTable([])
        releasedFormTable.addMouseListener(new MouseDoubleclickListener( { downloadReleasedFormXML() } ))

        populateDraftNewFormTable([])
        draftNewFormTable.addMouseListener(new MouseDoubleclickListener( { downloadDraftNewFormXML() } ))

        populateCdeReviewTable([])
        cdeReviewTable.addMouseListener(new MouseDoubleclickListener( { downloadCdeReviewXML() } ))

        this.loadMetadataConsoleProperties()

        if(removeCdeReviewTab)
        {
            tabbedPane.removeTabAt(2);
        }

        if(removeDraftNewFormTab)
        {
            tabbedPane.removeTabAt(1);
        }

        frame.show()
    }
}